#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

# Note: error_messages must be declared in prerequisites.sh

function parse_hardware_validation_args() {
  local machine_installation_type
  local product_installation_type
  local is_multi_node_installation
  local profile

  machine_installation_type="single_node"
  product_installation_type="core_platform"
  profile="$(get_mandatory_json_value ".profile")"
  if [[ "${profile}" == "ha" ]];
  then
    is_multi_node_installation="true"
  else
    is_multi_node_installation="false"
  fi

  if [[ "${is_multi_node_installation}" == "true" ]]; then
      machine_installation_type="multi_node"
  fi

  if [[ "$(is_full_installation)" == "true" ]]; then
      product_installation_type="full_automation_suite"
  fi  
  
  MOUNT_POINT="$(get_mandatory_json_value ".infra.longhorn.mount")"
  MINIMUM_DATA_DISK_SIZE=$(get_mandatory_json_value ".infra.hardware_requirements.${machine_installation_type}.${product_installation_type}.data_disk")    
  if [[ "${INSTALLSTANDALONEAICENTER}" == "1" ]]; then
    [[ "${is_multi_node_installation}" == "true" ]] && MINIMUM_CPU_CORES=8 || MINIMUM_CPU_CORES=16
  else
    MINIMUM_CPU_CORES=$(get_mandatory_json_value ".infra.hardware_requirements.${machine_installation_type}.${product_installation_type}.cpu")
  fi
  MINIMUM_MEMORY_IN_GB=$(get_mandatory_json_value ".infra.hardware_requirements.${machine_installation_type}.${product_installation_type}.memory")
  MINIMUM_RANCHER_DIR_SIZE=$(get_mandatory_json_value ".infra.os_disk_requirements.rancher_dir_size")
  MINIMUM_KUBELET_DIR_SIZE=$(get_mandatory_json_value ".infra.os_disk_requirements.kubelet_dir_size")
  MINIMUM_ETCD_DIR_SIZE=$(get_mandatory_json_value ".infra.os_disk_requirements.etcd_dir_size")
  
  # Hardware requirement based on NodeType
  MINIMUM_CPU_CORES_TASKMINING=$(get_mandatory_json_value ".infra.hardware_requirements.task_mining.cpu")
  MINIMUM_CPU_CORES_GPU=$(get_mandatory_json_value ".infra.hardware_requirements.gpu.cpu")
  MINIMUM_MEMORY_IN_GB_TASKMINING=$(get_mandatory_json_value ".infra.hardware_requirements.task_mining.memory")
  MINIMUM_MEMORY_IN_GB_GPU=$(get_mandatory_json_value ".infra.hardware_requirements.gpu.memory")

  # Reduce 10% reserved size
  MINIMUM_MEMORY_IN_GB_EXCLUDING_RESERVED=$(((MINIMUM_MEMORY_IN_GB*9)/10))
  MINIMUM_RANCHER_DIR_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_RANCHER_DIR_SIZE*9)/10))
  MINIMUM_KUBELET_DIR_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_KUBELET_DIR_SIZE*9)/10))
  MINIMUM_ETCD_DIR_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_ETCD_DIR_SIZE*9)/10))
  MINIMUM_DATA_DISK_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_DATA_DISK_SIZE*9)/10))    
  MINIMUM_MEMORY_IN_GB_TASKMINING_EXCLUDING_RESERVED=$(((MINIMUM_MEMORY_IN_GB_TASKMINING*9)/10))
  MINIMUM_MEMORY_IN_GB_GPU_EXCLUDING_RESERVED=$(((MINIMUM_MEMORY_IN_GB_GPU*9)/10))
}

function cpu_validation() {
  local error_messages=()
  local node_type=$1
  local cpu_count
  cpu_count=$(grep -c processor /proc/cpuinfo)
  if [[ -z $cpu_count ]]; then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.cpu_info_fetch_error" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".hardware_validation.cpu_info_fetch_error" "${MESSAGE_FILE}" )" 
    error_messages+=("Fetching system cpu info failed.")
  fi

  minimum_cpu_cores_required="${MINIMUM_CPU_CORES}"
  
  if [[ "${node_type}" == "task-mining" ]]; then
    minimum_cpu_cores_required="${MINIMUM_CPU_CORES_TASKMINING}"
  elif [[ "${node_type}" == "gpu" ]]; then
    minimum_cpu_cores_required="${MINIMUM_CPU_CORES_GPU}"
  fi

  # Stop installation if cpu count is less than 'minimum_cores' mentioned in defaults.json
  if (( cpu_count < minimum_cpu_cores_required )); then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.cpu_validation_error" "${MESSAGE_FILE}")" "$(get_readiness_mitigation ".hardware_validation.cpu_validation_error" "${MESSAGE_FILE}" "${minimum_cpu_cores_required}" "${cpu_count}")" 
    error_messages+=("Minimum CPU core requirement is $minimum_cpu_cores_required")
  fi

  if [[ ${#error_messages[@]} == 0 ]]; then
    add_test_result "PASSED" "$(get_readiness_message ".hardware_validation.cpu_validation_info" "${MESSAGE_FILE}" )"
  fi
}

function memory_validation() {
  local node_type=$1
  local memory_info
  local error_messages=()
  memory_info=$( grep MemTotal /proc/meminfo )
  if [[ -z "${memory_info}" ]]; then
    error_messages+=("Fetching system memory info failed.")
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.memory_info_fetch_error" )" "$(get_readiness_mitigation ".hardware_validation.memory_info_fetch_error" "${MESSAGE_FILE}" )" 
  fi

  # parse and get the memory in kb
  local memory_in_kb
  memory_in_kb=$(echo "${memory_info}" | xargs | cut -d' ' -f2)

  # convert memory info from kb to gb
  local memory_in_Gb=$(( "${memory_in_kb}"/1024/1024 ))

  minimum_memory_in_gb_required="${MINIMUM_MEMORY_IN_GB_EXCLUDING_RESERVED}"
  
  if [[ "${node_type}" == "task-mining" ]]; then
    minimum_memory_in_gb_required="${MINIMUM_MEMORY_IN_GB_TASKMINING_EXCLUDING_RESERVED}"
  elif [[ "${node_type}" == "gpu" ]]; then
    minimum_memory_in_gb_required="${MINIMUM_MEMORY_IN_GB_GPU_EXCLUDING_RESERVED}"
  fi

  # Stop installation if memory size is less 'than minimum_memory_size' GB mentioned in defaults.json
  if (( memory_in_Gb < minimum_memory_in_gb_required )); then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.memory_validation_error" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".hardware_validation.memory_validation_error" "${MESSAGE_FILE}" "${minimum_memory_in_gb_required}" "${memory_in_Gb}")" 
    error_messages+=("Minimum memory requirement is $minimum_memory_in_gb_required GiB")
  fi

  if [[ ${#error_messages[@]} == 0 ]]; then
    add_test_result "PASSED" "$(get_readiness_message ".hardware_validation.memory_validation_info" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".hardware_validation.memory_validation_info" "${MESSAGE_FILE}" )" 
  fi
}

function validate_fstab_entry() {
  local MOUNT_POINT=$1
  # fstab validation
  local mount_available_in_fstab
  mount_available_in_fstab=$(grep -w "${MOUNT_POINT}" /etc/fstab) || true
  if [[ -z "${mount_available_in_fstab}" ]]; then
    error_messages+=("$(get_readiness_message ".hardware_validation.fstab_missing_entry_error" "${MESSAGE_FILE}" "${MOUNT_POINT}" )")
    return
  fi

  # If it's getting executed as non-root user, the check does not work
  [[ "$(id -un 2>/dev/null || true)" != "root" ]] && return

  local validata_disk_mount
  validata_disk_mount=$(mount -fav|grep -w "${MOUNT_POINT}") || true
  if [[ -z "${validata_disk_mount}" ]]; then
    error_messages+=("$(get_readiness_message ".hardware_validation.fstab_invalid_entry_error" "${MESSAGE_FILE}" "${MOUNT_POINT}" )")
  fi
}

function data_disk_validation() {
  [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]] && add_test_result "INFO" "$(get_readiness_message ".hardware_validation.data_disk_validation_info" "${MESSAGE_FILE}" )" && return
  
  echo -e "$(get_readiness_message ".hardware_validation.data_disk_validation_start" "${MESSAGE_FILE}")"
  if [[ ! -d "${MOUNT_POINT}" ]]; then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.data_disk_attachment_error" "${MESSAGE_FILE}" "${MOUNT_POINT}" )" "$(get_readiness_mitigation ".hardware_validation.data_disk_attachment_error" "${MESSAGE_FILE}" )" 
    echo -e "$(get_readiness_message ".hardware_validation.data_disk_validation_end" "${MESSAGE_FILE}")\n"
    return
  fi

  local disk_info
  local disk_info_list
  disk_info=$(df -T "${MOUNT_POINT}" | tail -n +2 |  awk '{ print $2, $3 }')
  IFS=" " read -r -a disk_info_list <<< "${disk_info}"
  local type
  local size_in_Kb
  type="${disk_info_list[0]}"
  size_in_Kb="${disk_info_list[1]}"
  local size_in_GB=$((size_in_Kb/1024/1024))
  
  error_messages=()
  if [[ "${type}" != "ext4" && "${type}" != "xfs" ]]; then
    error_messages+=("$(get_readiness_message ".hardware_validation.data_disk_unsupported_format_error" "${MESSAGE_FILE}" "${type}" "${MOUNT_POINT}" )")
  fi

  if [[ "${size_in_GB}" -lt "${MINIMUM_DATA_DISK_SIZE_EXCLUDING_RESERVED}" ]]; then
    error_messages+=("$(get_readiness_message ".hardware_validation.disk_size_error" "${MESSAGE_FILE}" "${MOUNT_POINT}" "${MINIMUM_DATA_DISK_SIZE}" "${size_in_GB}")")
  fi

  validate_fstab_entry "${MOUNT_POINT}"
  if [[ "${#error_messages[@]}" != 0 ]];
  then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${MOUNT_POINT}")." "$(get_mitigation_string_from_array "${error_messages[@]}")"
  else
    add_test_result "PASSED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${MOUNT_POINT}")."
  fi
  echo -e "$(get_readiness_message ".hardware_validation.data_disk_validation_end" "${MESSAGE_FILE}")\n"
}

function rke2_disk_validation() {
  local rancher_partition
  local kubelet_partition
  local etcd_partition
  local rancher_path="/var/lib/rancher"
  local kubelet_path="/var/lib/kubelet"
  local etcd_path="/var/lib/rancher/rke2/server/db"
  local NODE_TYPE=$1

  if [[ "${NODE_TYPE}" == "none" || "${NODE_TYPE}" == "server" ]]; then 
    if [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]]; then
      add_test_result "INFO" "$(get_readiness_message ".hardware_validation.rke2_cluster_running_info" "${MESSAGE_FILE}" )"
      return
    fi
  else
    if [[ "$(systemctl is-enabled rke2-agent 2>/dev/null)" == "enabled" ]]; then
      add_test_result "INFO" "$(get_readiness_message ".hardware_validation.rke2_cluster_running_info" "${MESSAGE_FILE}" )"
      return
    fi
  fi

  echo -e "$(get_readiness_message ".hardware_validation.cluster_disk_validation_start" "${MESSAGE_FILE}")"
  error_messages=()
  # Check if /var/lib/rancher is mount with a separate device
  if [[ -d "${rancher_path}" ]]; then
    rancher_partition=$(df "${rancher_path}" | sed -n '2 p' | cut -d' ' -f1)
    # Check that the corresponding device is indeed mounted at the desired location
    [[ $(df ${rancher_path} --output=target | grep ${rancher_path}) != "${rancher_path}" ]] && error_messages+=("$(get_readiness_message ".hardware_validation.disk_mount_error" "${MESSAGE_FILE}" "${rancher_path}" )")
  else
    error_messages+=("$(get_readiness_message ".hardware_validation.disk_mount_error" "${MESSAGE_FILE}" "${rancher_path}" )")
  fi

  if [[ -n "$rancher_partition" && "$(get_size_in_GB "$rancher_partition")" -lt "${MINIMUM_RANCHER_DIR_SIZE_EXCLUDING_RESERVED}" ]]; then
    rancher_size=$(get_size_in_GB "$rancher_partition")
    error_messages+=("$(get_readiness_message ".hardware_validation.disk_size_error" "${MESSAGE_FILE}" "${rancher_partition}" "${MINIMUM_RANCHER_DIR_SIZE}" "${rancher_size}")")
  fi

  validate_fstab_entry "$rancher_path"

  if [[ "${#error_messages[@]}" != 0 ]];
  then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${rancher_path}")" "$(get_mitigation_string_from_array "${error_messages[@]}")"
  else
    add_test_result "PASSED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${rancher_path}")."
  fi

  error_messages=()
  # Check if /var/lib/kubelet is mount with a separate device
  if [[ -d "${kubelet_path}" ]]; then
    kubelet_partition=$(df "${kubelet_path}" | sed -n '2 p' | cut -d' ' -f1)
    # Check that the corresponding device is indeed mounted at the desired location
    [[ $(df ${kubelet_path} --output=target | grep ${kubelet_path}) != "${kubelet_path}" ]] && error_messages+=("$(get_readiness_message ".hardware_validation.disk_mount_error" "${MESSAGE_FILE}" "${kubelet_path}" )")
  else
    error_messages+=("$(get_readiness_message ".hardware_validation.disk_mount_error" "${MESSAGE_FILE}" "${kubelet_path}" )")
  fi

  if [[ -n "$kubelet_partition" && "$(get_size_in_GB "$kubelet_partition")" -lt "${MINIMUM_KUBELET_DIR_SIZE_EXCLUDING_RESERVED}" ]]; then
    kubelet_size=$(get_size_in_GB "$kubelet_partition")
    error_messages+=("$(get_readiness_message ".hardware_validation.disk_size_error" "${MESSAGE_FILE}" "${kubelet_partition}" "${MINIMUM_KUBELET_DIR_SIZE}" "${kubelet_size}")")
  fi

  validate_fstab_entry "$kubelet_path"
  if [[ "${#error_messages[@]}" != 0 ]];
  then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${kubelet_path}")" "$(get_mitigation_string_from_array "${error_messages[@]}")"
  else
    add_test_result "PASSED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${kubelet_path}")"
  fi
  echo -e "$(get_readiness_message ".hardware_validation.cluster_disk_validation_end" "${MESSAGE_FILE}")\n"
  
  echo -e "$(get_readiness_message ".hardware_validation.etcd_disk_validation_start" "${MESSAGE_FILE}")"
  error_messages=()
  if [[ "${NODE_TYPE}" == "none" || "${NODE_TYPE}" == "server" ]]; then 
    if [[ -d "${etcd_path}" ]]; then
      etcd_partition=$(df "${etcd_path}" | sed -n '2 p' | cut -d' ' -f1)
      if [[ "${etcd_partition}" == "${rancher_partition}" || "${etcd_partition}" == "${kubelet_partition}" ]]; then
        error_messages+=("$(get_readiness_message ".hardware_validation.etcd_disk_missing_mount_error" "${MESSAGE_FILE}" "${etcd_path}" )")
      fi

      if [[ "$(get_size_in_GB "$etcd_partition")" -lt "${MINIMUM_ETCD_DIR_SIZE_EXCLUDING_RESERVED}" ]]; then
        etcd_size=$(get_size_in_GB "$etcd_partition")
        error_messages+=("$(get_readiness_message ".hardware_validation.disk_size_error" "${MESSAGE_FILE}" "${etcd_partition}" "${MINIMUM_ETCD_DIR_SIZE}" "${etcd_size}")")
      fi
      
      validate_fstab_entry "$etcd_path"    
    else
      error_messages+=("$(get_readiness_message ".hardware_validation.etcd_disk_missing_mount_error" "${MESSAGE_FILE}" "${etcd_path}" )")
    fi
  fi 
  if [[ "${#error_messages[@]}" != 0 ]];
  then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${etcd_path}")" "$(get_mitigation_string_from_array "${error_messages[@]}")"
  else
    add_test_result "PASSED" "$(get_readiness_message ".hardware_validation.disk_mount_info" "${MESSAGE_FILE}" "${etcd_path}")"
  fi
  echo -e "$(get_readiness_message ".hardware_validation.etcd_disk_validation_end" "${MESSAGE_FILE}")\n"
}

# given the device name, returns the size in GB
function get_size_in_GB() {
  local size_in_bytes
  size_in_bytes=$(lsblk "$1" -b --nodeps --noheadings --paths --output size)
  local divisor=$((1024*1024*1024))
  local size_in_GB=$(((size_in_bytes+divisor-1)/divisor)) # this will round up value
  echo ${size_in_GB}
}

# Given a directory path, it will check if noexec option is set on the corresponding mount point
function noexec_flag_on_mountpoint() {
  local directory=$1
  mount_options=$(findmnt -T "${directory}" -o OPTIONS -n)
  if echo "$mount_options" | grep -q 'noexec';
  then
    add_test_result "FAILED" "$(get_readiness_message ".hardware_validation.noexec_flag_error" "${MESSAGE_FILE}" "${directory}")" "$(get_readiness_mitigation ".hardware_validation.noexec_flag_error" "${MESSAGE_FILE}" "${directory}")" 
  else
    add_test_result "PASSED" "$(get_readiness_message ".hardware_validation.noexec_flag_info" "${MESSAGE_FILE}" "${directory}")"
  fi
}

function validate_noexec_option() {
  echo -e "\n$(get_readiness_message ".hardware_validation.noexec_flag_verification_start" "${MESSAGE_FILE}")"
  noexec_flag_on_mountpoint "/var/lib/longhorn"
  noexec_flag_on_mountpoint "/var/lib/kubelet"
  noexec_flag_on_mountpoint "/var/lib/rancher"
  echo -e "$(get_readiness_message ".hardware_validation.noexec_flag_verification_end" "${MESSAGE_FILE}")\n"
}