#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${SCRIPT_DIR}/installUiPathAS.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${SCRIPT_DIR}/installUiPathAS.sh"
  set +eu
else
  echo "Readiness configuration script not found at ${SCRIPT_DIR}/installUiPathAS.sh"
  exit 1
fi

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

if [[ -f "${MODULES_DIR}/offline-utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/offline-utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/offline-utils.sh"
  exit 1
fi

if [[ -f "${SCRIPT_DIR}/defaults.json" ]] ; then
  export DEFAULTSFILE="${SCRIPT_DIR}/defaults.json"
else
  error "Default json file not found at ${SCRIPT_DIR}/defaults.json"
fi

export MERGEDINPUTFILE="${SCRIPT_DIR}/merged_input.json"


CHECKS_DIR=$(cd "${SCRIPT_DIR}/Prerequisite_Checks" && pwd)
if [[ -z "${CHECKS_DIR}" ]]; then
  error "Could not determine Prerequisite_Checks path"
fi

MESSAGE_FILE="${CHECKS_DIR}/readiness_message_en.json"
if [[ ! -f "${MESSAGE_FILE}" ]] ; then
  echo "Readiness message file not present"
  exit 1
fi


export PATH="$PATH:/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/bin"
export KUBECONFIG="${KUBECONFIG_PATH:-/etc/rancher/rke2/rke2.yaml}"

ENABLED_SEVICES=""
export MITIGATION_DELIMITER="\n    "
UNEXPECTED_FAILURE="false"
PREREQUISITE_CHECK_FAILED="false"
INFILE="/opt/UiPathAutomationSuite/cluster_config.json"
NODES=""
NODE_TYPE=""
INSTALL_TYPE=""
PRIMARY_SERVER="false"
INSTALLSTANDALONEAICENTER="0"
READINESS_BUNDLE_PATH=""
READINESS_TMP_BUNDLE_PATH="/opt/UiPathAutomationSuite"
SKIP_BUNDLE_EXTRACTION="0"
DISABLED_PREREQ_CHECK_LIST=""
# if this flag is true, this will skill podman install, bundle extraction, image load
SKIP_BUNDLE_VALIDATION="0"

required_infra_pkgs=(iscsi-initiator-utils nfs-utils rpcbind uuidgen nc openssl httpd-tools)
required_offline_pkgs=(podman)
required_online_packages=(wget unzip)
load_os_utils "${MODULES_DIR}" &>/dev/null

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "validateUiPathASReadiness.sh Runs pre-install validation checks"
  echo 
  echo "Arguments:"
  echo " -i|--input                                             Specify cluster config json file"
  echo " -m|--machines                                          Comma seperated IP address / hostnames of all the machines"
  echo " -n|--node-type                                         Possible values [server|agent|taskmining|gpu]. Defaults to server"
  echo " --install-type                                         Possible values [online|offline]. Defaults to online"
  echo " --install-standalone-aicenter                          Run checks for standalone aicenter service only"
  echo " --prereq-bundle </uipath/offline_bundle.tar.gz>        Specify the path to the prereq bundle"
  echo " --prereq-tmp-folder </uipath>                          Specify parent directory for the temporary install folder"
  echo " -h|--help                                              Display help"
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  local iter
  while (("$#")); do
    # export COMPONENT=$1
    case "$1" in
    -d | --debug)
      shift
      set -x
    ;;

    -h | --help)
      display_usage
      shift
      exit 0
    ;;

    -i | --input)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        INFILE=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;    

    -m | --machines)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODES="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;;

    -n | --node-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODE_TYPE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;; 

    --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        INSTALL_TYPE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;;

    -p | --primary-server)
      PRIMARY_SERVER="true"
      shift
    ;;  

    --install-standalone-aicenter)
      shift
      INSTALLSTANDALONEAICENTER="1"
    ;;

    --prereq-bundle)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        READINESS_BUNDLE_PATH=$ABSOLUTEPATH
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;;

    --prereq-tmp-folder)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        READINESS_TMP_BUNDLE_PATH=$ABSOLUTEPATH
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;;

    --skip-bundle-extraction)
      SKIP_BUNDLE_EXTRACTION=1
      shift
      ;;

    --skip-bundle-validation)
      SKIP_BUNDLE_VALIDATION=1
      shift
      ;;

    --disabled-prereq-check)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        DISABLED_PREREQ_CHECK_LIST="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      echo "Error: Unsupported flag $1"
      exit 1
      ;;
    esac
  done
  # display_usage
  # check if provided input file was provided and exists else display option to create config file
  if [[ -z "${INFILE}" || ! -f "${INFILE}" ]]; then
    create_readiness_config "${INSTALL_TYPE}" "${INFILE}"
  fi

  if [[ ! -s "${INFILE}" ]]; then
    error "Readiness input file ${INFILE} is empty."
  else
    echo -e "\u001b[32mUsing readiness input file ${INFILE}\e[0m"
  fi

  ##Validate Ips
  local invalid_ips=()
  if [[ -n "${NODES}" ]]; then
    NODES=$(echo "${NODES}" | tr ',' '\n')
    while IFS=$'\n' read -r iter
    do
      iter=$(printf '%s' "$iter" | tr -d '[:space:]')
      [[ -z "${iter}" ]] && continue
      if ! validate_ipaddress "${iter}" >> /dev/null 2>&1 ; then
        invalid_ips+=("Invalid IP: ${iter}")
      fi
    done <<< "${NODES}"

    if [[ ${#invalid_ips[@]} != 0 ]] ; then
      failure_output "Invalid IP(s) provided in input" "${invalid_ips[@]}"
    fi
  fi

  # Validate node type
  if [[ -z "${NODE_TYPE}" ]] ; then
    NODE_TYPE="server"
  else
    case "${NODE_TYPE}" in
      server)
      ;;
      agent)
      ;;
      task-mining)
      ;;
      gpu)
      ;;
      *)
      error "Unsupported node type: ${NODE_TYPE} provided"
      ;;
    esac
  fi

  # Validate install type
  if [[ -z "${INSTALL_TYPE}" ]] ; then
    INSTALL_TYPE="online"
  else
    case "$INSTALL_TYPE" in
      online)
      ;;
      offline)
      ;;
      *)
      error "Unsupported install type: ${INSTALL_TYPE} provided"
      ;;
    esac
  fi

  if [[ -z "${PRIMARY_SERVER}" ]] ; then
    PRIMARY_SERVER="false"
  fi

  if [[ -z "${INSTALLSTANDALONEAICENTER}" ]] ; then
    INSTALLSTANDALONEAICENTER="0"
  fi

  if [[ "${INSTALL_TYPE}" == "offline" && -z ${READINESS_BUNDLE_PATH} && "${SKIP_BUNDLE_VALIDATION}" == "0" ]] ; then
    error_with_code "10030"
  fi

  if [[ -n "${SKIP_VALIDATION_NAMES}" ]] ; then
    error_with_code "10030"
  fi

  export INSTALLSTANDALONEAICENTER
}

function is_failure() {
  local exitCode="$1"
  case "$exitCode" in
    0)
    ;;
    51)
      PREREQUISITE_CHECK_FAILED="true"
    ;;
    *)
      UNEXPECTED_FAILURE="true"
      error_without_exit "Unexpected error occurred while executing $2 check"
    ;;
  esac
}

function is_action_disabled() {
  local actionName="$1"
  local result
  result=0
  if [[ -n "${DISABLED_PREREQ_CHECK_LIST}" && "${DISABLED_PREREQ_CHECK_LIST}" == *"${actionName}"* ]]; then
    result=1
  fi
  echo ${result}
}

function run_all() {
  [[ $- == *"e"* ]] && set +e
  local exitCode version

  # Fetch compatible installer version
  version="$(get_mandatory_json_value ".version")"

  if [[ $(is_action_disabled "proxy") == "0" ]]; then
    "${CHECKS_DIR}/proxy/main.sh" "$@"
    is_failure "$?" "proxy"
  fi

  if [[ $(is_action_disabled "connectivity") == "0" ]]; then
    "${CHECKS_DIR}/connectivity/main.sh" "$@" --enabled-services "${ENABLED_SEVICES}"
    is_failure "$?" "connectivity"
  fi

  if [[ $(is_action_disabled "certificates") == "0" ]]; then
    "${CHECKS_DIR}/certificates/main.sh" "$@"
    is_failure "$?" "certificates"
  fi

  if [[ $(is_action_disabled "hardware-validation") == "0" ]]; then
    "${CHECKS_DIR}/hardware-validation/main.sh" "$@"
    is_failure "$?" "hardware-validation"
  fi

  if [[ $(is_action_disabled "systemd-services") == "0" ]]; then
    "${CHECKS_DIR}/systemd-services/main.sh" "$@"
    is_failure "$?" "systemd-services"
  fi

  if [[ $(is_action_disabled "sql") == "0" ]]; then
    "${CHECKS_DIR}/sql/main.sh" "$@" --enabled-services "${ENABLED_SEVICES}"
    is_failure "$?" "sql"
  fi

  ## Keep this at last
  if [[ $PREREQUISITE_CHECK_FAILED == "true" ]] || [[ $UNEXPECTED_FAILURE == "true" ]] ; then
    add_test_result "FAILED" "$(get_readiness_message ".common.readiness_check_failed" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".common.readiness_check_failed" "${MESSAGE_FILE}")"
  fi

  if [[ $UNEXPECTED_FAILURE == "false" ]] && [[ $PREREQUISITE_CHECK_FAILED == "false" ]] ; then
    add_test_result "PASSED" "$(get_readiness_message ".common.readiness_check_passed" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".common.readiness_check_passed" "${MESSAGE_FILE}" "${version}")"
    exitCode=0
  else
    exitCode=1
  fi
  exit $exitCode
}

function check_if_required_packages_available() {
  local error_messages=()
  local package
  local required_packages=("jq" "dig" "nc" "curl" "podman" "timeout")
  for package in "${required_packages[@]}"
  do
    if ! which "${package}" >> /dev/null 2>&1 ; then
      error_messages+=("'${package}' not found in PATH or is not installed. Make sure '${package}' is installed and is available in the PATH")
    fi
  done
  if [[ ${#error_messages[@]} != 0 ]] ; then
    failure_output "Required packages not available" "${error_messages[@]}"
  fi
}

function extract_readiness_artifacts() {
  if [[ "${SKIP_BUNDLE_EXTRACTION}" == "0" ]]; then
    echo "Running the readiness validation using ${READINESS_BUNDLE_PATH}"
    extract_offline_bundle "${READINESS_BUNDLE_PATH}" "${READINESS_TMP_BUNDLE_PATH}"
  else
    echo "Using already extracted readiness validation using ${READINESS_TMP_BUNDLE_PATH}"
  fi
}

function import_offline_readiness_artifacts() {
  if [[ $(is_action_disabled "sql") == "0" ]]; then
    local offline_readiness_install_folder="$1"
    local readiness_bundle="${offline_readiness_install_folder}/readiness-bundle.tar.gz"
    local readiness_sha="${offline_readiness_install_folder}/readiness-bundle-sha256sum.txt"
    local offline_readiness_dir="${offline_readiness_install_folder}/readiness"
    check_bundle "${readiness_bundle}" "${readiness_sha}"
    echo "Extracting ${readiness_bundle} to ${offline_readiness_dir}"
    mkdir -p "${offline_readiness_dir}" && tar -xzf "${readiness_bundle}" -C "${offline_readiness_dir}"
    rm -f "${readiness_bundle}"
    image_tar="${offline_readiness_dir}/images/images.tar"
    podman load -i "${image_tar}" || error_with_code "10017" "${image_tar}"
    rm -rf "${offline_readiness_dir}"
    rm -f "${readiness_sha}"
  fi
}

function check_infra_prerequisites() {
    create_missing_packages_and_commands_list "${required_infra_pkgs[@]}"
    if [ -n "$MISSING_PKGS" ]; then
      add_test_result "WARN" "$(get_readiness_message ".infra.missing_infra_pkg_error" "${MESSAGE_FILE}" "${MISSING_PKGS}" )" "$(get_readiness_mitigation ".infra.missing_infra_pkg_error" "${MESSAGE_FILE}" )"
    else
      add_test_result "PASSED" "All required packages are present: ${required_infra_pkgs[*]}"
    fi
    
    if [[ "${INSTALL_TYPE}" == "offline" ]]; then
      return
    fi

    unset MISSING_PKGS
    create_missing_packages_and_commands_list "${required_online_packages[@]}"
    if [ -n "$MISSING_PKGS" ]; then
      add_test_result "FAILED" "$(get_readiness_message ".infra.missing_online_pkg_error" "${MESSAGE_FILE}" "${MISSING_PKGS}" )" "$(get_readiness_mitigation ".infra.missing_online_pkg_error" "${MESSAGE_FILE}" )"
    else
      add_test_result "PASSED" "All required packages are present: ${required_online_packages[*]}"
    fi
}

function check_infra_prerequisites_and_exit() {
    check_required_packages_and_commands "${required_infra_pkgs[@]}"
}

function check_offline_prerequisites() {
    create_missing_packages_and_commands_list "${required_offline_pkgs[@]}"
    if [ -n "$MISSING_PKGS" ]; then
      add_test_result "WARN" "$(get_readiness_message ".infra.missing_offline_pkg_error" "${MESSAGE_FILE}" "${MISSING_PKGS}" )" "$(get_readiness_mitigation ".infra.missing_offline_pkg_error" "${MESSAGE_FILE}" )"
    else
      add_test_result "PASSED" "Offline Install - All required packages are present: ${required_offline_pkgs[*]}"
    fi
}

function check_offline_prerequisites_and_exit() {
    check_required_packages_and_commands "${required_offline_pkgs[@]}"
}

function validate_installed_software() {
  print_center "Software Components Validation"
  check_infra_prerequisites
  [[ $INSTALL_TYPE == "offline" ]] && check_offline_prerequisites
  echo ""
  print_center ""
}

# We we install podman, it bydefault create podman network and store metadata at this localtion /etc/cni/. While uninstalling cluster, 
# rke2 unstall script clean all data available at location /etc/cni/*. So we need to check and create default network again as podmand 
# don't create default network post installation https://github.com/containers/podman/issues/12651

function validate_podman_default_network() {
  local network_list
  if [[ $(is_action_disabled "sql") == "0" ]]; then 
    network_list=$(podman network ls --quiet)
    if [[ -z "${network_list}" || "${network_list}" == " " ]]; then
      podman network create podman 2>/dev/null || true
    fi
  fi
}

function configure_readiness_prereq() {
  local podman_folder

  load_prerequisites_package_utils "${MODULES_DIR}"

  if [[ -n "${READINESS_BUNDLE_PATH}" ]]; then
    extract_readiness_artifacts 
    install_offline_readiness_validation_prereqs "${READINESS_TMP_BUNDLE_PATH}"
  fi
  if [[ "${INSTALL_TYPE}" == "offline" ]]; then
    podman_folder="$(get_mandatory_json_value ".infra.offline.dir")/lib/containers"
  else
    [[ -z "${READINESS_BUNDLE_PATH}" ]] && install_online_readiness_validation_prereqs
    podman_folder="$(get_json_value_or_return_default_value ".infra.online.dir" "/uipath")/lib/containers"
  fi
  configure_podman "${podman_folder}"

  # inline ENV for https://github.com/containers/podman/pull/5412 and https://github.com/containers/podman/issues/5960
  export TMPDIR="${READINESS_TMP_BUNDLE_PATH}/tmp"
  mkdir -p "${TMPDIR}"

  [[ "${INSTALL_TYPE}" == "offline" && -n "${READINESS_BUNDLE_PATH}" ]] && import_offline_readiness_artifacts "${READINESS_TMP_BUNDLE_PATH}"

  validate_podman_default_network
}

function main() {
  parse_long_args "$@"

  diff "$INFILE" "${MERGEDINPUTFILE}" >> /dev/null 2>&1 || merge_default_values
  export INFILE
  ENABLED_SEVICES=$(get_enabled_services)

  if [[ $(is_action_disabled "sql") == "0" ]]; then
    configure_readiness_prereq
    check_if_required_packages_available
  fi

  validate_installed_software
  run_all "$@"
}

main "$@"